//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2023 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef f_AT_TRACEFILEENCODING_H
#define f_AT_TRACEFILEENCODING_H

#include <vd2/system/vdtypes.h>
#include <vd2/system/refcount.h>
#include <at/atcore/snapshotimpl.h>

class ATSaveStateMemoryBuffer;

class IATTraceFmtCodec : public IVDRefUnknown {
protected:
	~IATTraceFmtCodec() = default;

public:
	static constexpr uint32 kTypeID = "IATTraceFmtCodec"_vdtypeid;

	virtual void Validate(uint32 rowSize) const = 0;
	virtual void Encode(ATSaveStateMemoryBuffer& dst, const uint8 *src, uint32 rowSize, size_t rowCount) const = 0;
	virtual void Decode(const ATSaveStateMemoryBuffer& src, uint8 *dst, uint32 rowSize, size_t rowCount) const = 0;
};

class ATSavedTraceCodecNull final : public ATSnapExchangeObject<ATSavedTraceCodecNull, "ATSavedTraceCodecNull", IATTraceFmtCodec> {
public:
	template<ATExchanger T>
	void Exchange(T& ex) {
	}

	void Validate(uint32 rowSize) const override {}
	void Encode(ATSaveStateMemoryBuffer& dst, const uint8 *src, uint32 rowSize, size_t rowCount) const override;
	void Decode(const ATSaveStateMemoryBuffer& src, uint8 *dst, uint32 rowSize, size_t rowCount) const override;
};

class ATSavedTraceCodecSparse final : public ATSnapExchangeObject<ATSavedTraceCodecSparse, "ATSavedTraceCodecSparse", IATTraceFmtCodec> {
public:
	template<ATExchanger T>
	void Exchange(T& ex) {
	}

	void Validate(uint32 rowSize) const override;
	void Encode(ATSaveStateMemoryBuffer& dst, const uint8 *src, uint32 rowSize, size_t rowCount) const override;
	void Decode(const ATSaveStateMemoryBuffer& src, uint8 *dst, uint32 rowSize, size_t rowCount) const override;
};

////////////////////////////////////////////////////////////////////////////////

class IATTraceFmtPredictor : public IVDRefUnknown {
protected:
	~IATTraceFmtPredictor() = default;

public:
	static constexpr uint32 kTypeID = "IATTraceFmtPredictor"_vdtypeid;

	virtual void Validate(uint32 rowSize) const = 0;
	virtual void Reset() = 0;
	virtual void Encode(uint8 *dst, uint32 rowSize, size_t rowCount) = 0;
	virtual void Decode(uint8 *dst, uint32 rowSize, size_t rowCount) = 0;
};

class ATSavedTracePredictorXOR final : public ATSnapExchangeObject<ATSavedTracePredictorXOR, "ATSavedTracePredictorXOR", IATTraceFmtPredictor> {
public:
	ATSavedTracePredictorXOR() = default;
	ATSavedTracePredictorXOR(uint32 offset, uint32 size);

	template<ATExchanger T_Ex>
	void Exchange(T_Ex& ex) {
		ex.Transfer("offset", &mXorOffset);
		ex.Transfer("size", &mXorSize);
	}

	void Validate(uint32 rowSize) const override;
	void Reset() override;
	void Encode(uint8 *dst, uint32 rowSize, size_t rowCount) override;
	void Decode(uint8 *dst, uint32 rowSize, size_t rowCount) override;

private:
	uint32 mXorOffset = 0;
	uint32 mXorSize = 0;
	uint8 mPredBuf[32] {};
};

class ATSavedTracePredictorEA final : public ATSnapExchangeObject<ATSavedTracePredictorEA, "ATSavedTracePredictorEA", IATTraceFmtPredictor> {
public:
	ATSavedTracePredictorEA() = default;
	ATSavedTracePredictorEA(uint32 offset);

	template<ATExchanger T_Ex>
	void Exchange(T_Ex& ex) {
		ex.Transfer("offset", &mOffset);
	}

	void Validate(uint32 rowSize) const override;
	void Reset() override;
	void Encode(uint8 *dst, uint32 rowSize, size_t rowCount) override;
	void Decode(uint8 *dst, uint32 rowSize, size_t rowCount) override;

private:
	uint32 mOffset = 0;
	uint32 mPrev = 0;
};

class ATSavedTracePredictorPC final : public ATSnapExchangeObject<ATSavedTracePredictorPC, "ATSavedTracePredictorPC", IATTraceFmtPredictor> {
public:
	ATSavedTracePredictorPC() = default;
	ATSavedTracePredictorPC(uint32 offset);

	template<ATExchanger T_Ex>
	void Exchange(T_Ex& ex) {
		ex.Transfer("offset", &mOffset);
	}

	void Validate(uint32 rowSize) const override;
	void Reset() override;
	void Encode(uint8 *dst, uint32 rowSize, size_t rowCount) override;
	void Decode(uint8 *dst, uint32 rowSize, size_t rowCount) override;

private:
	uint32 mOffset = 0;
	uint16 mPrevPC = 0;

	uint16 mPredBuf[64*1024];
};

class ATSavedTracePredictorInsn final : public ATSnapExchangeObject<ATSavedTracePredictorInsn, "ATSavedTracePredictorInsn", IATTraceFmtPredictor> {
public:
	ATSavedTracePredictorInsn() = default;
	ATSavedTracePredictorInsn(uint32 insnOffset, uint32 insnSize, uint32 pcOffset, uint32 flagsBitOffset);

	template<ATExchanger T_Ex>
	void Exchange(T_Ex& ex) {
		ex.Transfer("pc_offset", &mPCOffset);
		ex.Transfer("insn_offset", &mInsnOffset);
		ex.Transfer("insn_size", &mInsnSize);
		ex.Transfer("flags_bit_offset", &mFlagsBitOffset);

		if constexpr(ex.IsReader) {
			if (mInsnSize == 0 || mInsnSize > 8)
				throw ATInvalidSaveStateException();

			if (8 - (mFlagsBitOffset * 7) < mInsnSize)
				throw ATInvalidSaveStateException();
		}
	}

	void Validate(uint32 rowSize) const override;
	void Reset() override;
	void Encode(uint8 *dst, uint32 rowSize, size_t rowCount) override;
	void Decode(uint8 *dst, uint32 rowSize, size_t rowCount) override;

private:
	uint32 mPCOffset = 0;
	uint32 mInsnOffset = 0;
	uint8 mInsnSize = 0;
	uint32 mFlagsBitOffset = 0;
	uint32 mFlagsOffset = 0;
	uint8 mFlagsShift = 0;

	uint8 mPrevFlags = 0;

	uint8 mPredBuf[64*1024];
};

class ATSavedTracePredictorHorizDelta32 final : public ATSnapExchangeObject<ATSavedTracePredictorHorizDelta32, "ATSavedTracePredictorHorizDelta32", IATTraceFmtPredictor> {
public:
	ATSavedTracePredictorHorizDelta32() = default;
	ATSavedTracePredictorHorizDelta32(uint32 dstOffset, uint32 predOffset);

	template<ATExchanger T_Ex>
	void Exchange(T_Ex& ex) {
		ex.Transfer("offset", &mDstOffset);
		ex.Transfer("pred_offset", &mPredOffset);
	}

	void Validate(uint32 rowSize) const override;
	void Reset() override;
	void Encode(uint8 *dst, uint32 rowSize, size_t rowCount) override;
	void Decode(uint8 *dst, uint32 rowSize, size_t rowCount) override;

private:
	uint32 mDstOffset = 0;
	uint32 mPredOffset = 0;
};

class ATSavedTracePredictorVertDelta32 final : public ATSnapExchangeObject<ATSavedTracePredictorVertDelta32, "ATSavedTracePredictorVertDelta32", IATTraceFmtPredictor> {
public:
	ATSavedTracePredictorVertDelta32() = default;
	ATSavedTracePredictorVertDelta32(uint32 offset, uint32 bias);

	template<ATExchanger T_Ex>
	void Exchange(T_Ex& ex) {
		ex.Transfer("offset", &mOffset);
		ex.Transfer("bias", &mBias);
	}

	void Validate(uint32 rowSize) const override;
	void Reset() override;
	void Encode(uint8 *dst, uint32 rowSize, size_t rowCount) override;
	void Decode(uint8 *dst, uint32 rowSize, size_t rowCount) override;

private:
	uint32 mOffset = 0;
	sint32 mBias = 0;
	uint32 mPrev = 0;
};

class ATSavedTracePredictorVertDelta8 final : public ATSnapExchangeObject<ATSavedTracePredictorVertDelta8, "ATSavedTracePredictorVertDelta8", IATTraceFmtPredictor> {
public:
	ATSavedTracePredictorVertDelta8() = default;
	ATSavedTracePredictorVertDelta8(uint32 offset, uint32 count);

	template<ATExchanger T_Ex>
	void Exchange(T_Ex& ex) {
		ex.Transfer("offset", &mOffset);
		ex.Transfer("count", &mCount);

		if constexpr(ex.IsReader) {
			if (!mCount)
				throw ATInvalidSaveStateException();
		}
	}

	void Validate(uint32 rowSize) const override;
	void Reset() override;
	void Encode(uint8 *dst, uint32 rowSize, size_t rowCount) override;
	void Decode(uint8 *dst, uint32 rowSize, size_t rowCount) override;

private:
	uint32 mOffset = 0;
	uint32 mCount = 0;
	uint8 mPredBuf[32] {};
};

class ATSavedTracePredictorDelta32TablePrev8 final : public ATSnapExchangeObject<ATSavedTracePredictorDelta32TablePrev8, "ATSavedTracePredictorDelta32TablePrev8", IATTraceFmtPredictor> {
public:
	ATSavedTracePredictorDelta32TablePrev8() = default;
	ATSavedTracePredictorDelta32TablePrev8(uint32 valueOffset, uint32 opcodeOffset);

	template<ATExchanger T_Ex>
	void Exchange(T_Ex& ex) {
		ex.Transfer("offset", &mValueOffset);
		ex.Transfer("op_offset", &mOpcodeOffset);
	}

	void Validate(uint32 rowSize) const override;
	void Reset() override;
	void Encode(uint8 *dst, uint32 rowSize, size_t rowCount) override;
	void Decode(uint8 *dst, uint32 rowSize, size_t rowCount) override;

private:
	uint32 mValueOffset = 0;
	uint32 mOpcodeOffset = 0;
	uint8 mPrevOp = 0;
	uint32 mPredBuf[256] {};
};

class ATSavedTracePredictorXor32Table8 final : public ATSnapExchangeObject<ATSavedTracePredictorXor32Table8, "ATSavedTracePredictorXor32Table8", IATTraceFmtPredictor> {
public:
	ATSavedTracePredictorXor32Table8() = default;
	ATSavedTracePredictorXor32Table8(uint32 valueOffset, uint32 opcodeOffset);

	template<ATExchanger T_Ex>
	void Exchange(T_Ex& ex) {
		ex.Transfer("offset", &mValueOffset);
		ex.Transfer("pred_offset", &mPredOffset);
	}

	void Validate(uint32 rowSize) const override;
	void Reset() override;
	void Encode(uint8 *dst, uint32 rowSize, size_t rowCount) override;
	void Decode(uint8 *dst, uint32 rowSize, size_t rowCount) override;

private:
	uint32 mValueOffset = 0;
	uint32 mPredOffset = 0;
	uint32 mPredBuf[256] {};
};

class ATSavedTracePredictorXor32TablePrev16 final : public ATSnapExchangeObject<ATSavedTracePredictorXor32TablePrev16, "ATSavedTracePredictorXor32TablePrev16", IATTraceFmtPredictor> {
public:
	ATSavedTracePredictorXor32TablePrev16() = default;
	ATSavedTracePredictorXor32TablePrev16(uint32 valueOffset, uint32 pcOffset);

	template<ATExchanger T_Ex>
	void Exchange(T_Ex& ex) {
		ex.Transfer("offset", &mValueOffset);
		ex.Transfer("pc_offset", &mPCOffset);
	}

	void Validate(uint32 rowSize) const override;
	void Reset() override;
	void Encode(uint8 *dst, uint32 rowSize, size_t rowCount) override;
	void Decode(uint8 *dst, uint32 rowSize, size_t rowCount) override;

private:
	uint32 mValueOffset = 0;
	uint32 mPCOffset = 0;
	uint16 mPrevPC = 0;
	uint32 mPredBuf[64*1024] {};
};

class ATSavedTracePredictorSignMag32 final : public ATSnapExchangeObject<ATSavedTracePredictorSignMag32, "ATSavedTracePredictorSignMag32", IATTraceFmtPredictor> {
public:
	ATSavedTracePredictorSignMag32() = default;
	ATSavedTracePredictorSignMag32(uint32 offset);

	template<ATExchanger T_Ex>
	void Exchange(T_Ex& ex) {
		ex.Transfer("offset", &mOffset);
	}

	void Validate(uint32 rowSize) const override;
	void Reset() override;
	void Encode(uint8 *dst, uint32 rowSize, size_t rowCount) override;
	void Decode(uint8 *dst, uint32 rowSize, size_t rowCount) override;

private:
	uint32 mOffset = 0;
};

#endif
