//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2023 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#include <stdafx.h>
#include <at/atcore/savestate.h>
#include <at/atcore/snapshotimpl.h>
#include <vd2/system/binary.h>

#include "tracefileencoding.h"

//#define AT_PROFILE_TRACE_CODEC_STATISTICS

void ATSavedTraceCodecNull::Encode(ATSaveStateMemoryBuffer& dst, const uint8 *src, uint32 rowSize, size_t rowCount) const {
	dst.mBuffer.assign(src, src + rowSize * rowCount);
}

void ATSavedTraceCodecNull::Decode(const ATSaveStateMemoryBuffer& src, uint8 *dst, uint32 rowSize, size_t rowCount) const {
	const size_t tableSize = rowSize * rowCount;

	if (src.mBuffer.size() != tableSize)
		throw ATInvalidSaveStateException();

	memcpy(dst, src.mBuffer.data(), tableSize);
}

void ATSavedTraceCodecSparse::Validate(uint32 rowSize) const {
	if (rowSize > 32)
		throw ATInvalidSaveStateException();
}

void ATSavedTraceCodecSparse::Encode(ATSaveStateMemoryBuffer& dst, const uint8 *src, uint32 rowSize, size_t rowCount) const {
	if (rowSize > 32)
		throw MyError("Cannot encode trace stripe: unsupported row geometry.");

	vdfastvector<uint8> codecData;
	uint8 codecbuf[36] {};
				
#ifdef AT_PROFILE_TRACE_CODEC_STATISTICS
	uint32 stat[32] {};
#endif

	for(uint32 i = 0; i < rowCount; ++i) {
		const uint8 *VDRESTRICT rowSrc = &src[i * rowSize];

		uint32 mask = 0;
		uint8 *dst = codecbuf + 4;

		for(uint32 j=0; j<rowSize; ++j) {
			if (rowSrc[j]) {
				*dst++ = rowSrc[j];
				mask |= (1 << j);
#ifdef AT_PROFILE_TRACE_CODEC_STATISTICS
				++stat[j];
#endif
			}
		}

		if (rowSize <= 8) {
			codecbuf[3] = (uint8)mask;
			codecData.insert(codecData.end(), codecbuf+3, dst);
		} else if (rowSize <= 16) {
			VDWriteUnalignedLEU16(codecbuf+2, (uint16)mask);
			codecData.insert(codecData.end(), codecbuf+2, dst);
		} else if (rowSize <= 24) {
			VDWriteUnalignedLEU32(codecbuf, mask << 8);
			codecData.insert(codecData.end(), codecbuf+1, dst);
		} else {
			VDWriteUnalignedLEU32(codecbuf, mask);
			codecData.insert(codecData.end(), codecbuf, dst);
		}
	}

#ifdef AT_PROFILE_TRACE_CODEC_STATISTICS
	int stat2[32];

	for(int i=0; i<32; ++i)
		stat2[i] = (stat[i] * 200 / 524288 + 1) / 2;

	VDDEBUG2("Change states: %3u %3u %3u %3u | %3u %3u %3u %3u | %3u %3u %3u %3u | %3u %3u %3u %3u | %3u %3u %3u %3u | %3u %3u %3u %3u | %3u %3u %3u %3u\n"
		, stat2[ 0], stat2[ 1], stat2[ 2], stat2[ 3]
		, stat2[ 4], stat2[ 5], stat2[ 6], stat2[ 7]
		, stat2[ 8], stat2[ 9], stat2[10], stat2[11]
		, stat2[12], stat2[13], stat2[14], stat2[15]
		, stat2[16], stat2[17], stat2[18], stat2[19]
		, stat2[20], stat2[21], stat2[22], stat2[23]
		, stat2[24], stat2[25], stat2[26], stat2[27]
	);
#endif

	dst.mBuffer.assign(codecData.begin(), codecData.end());
}

void ATSavedTraceCodecSparse::Decode(const ATSaveStateMemoryBuffer& buf, uint8 *dst, uint32 rowSize, size_t rowCount) const {
	uint8 srcTail[64] {};

	memset(dst, 0, rowSize * rowCount);

	const uint8 *src = buf.mBuffer.data();
	const uint8 *srcEnd = src + buf.mBuffer.size();
	for(uint32 row = 0; row < rowCount; ++row) {
		uint8 *VDRESTRICT decodeDst = &dst[rowSize*row];
		const uint8 *VDRESTRICT decodeSrc = src;

		if (srcEnd - src < 64) {
			memcpy(srcTail, src, srcEnd - src);
			decodeSrc = srcTail;
		}

		const uint8 *VDRESTRICT decodeSrc2 = decodeSrc;

		uint32 mask = 0;

		if (rowSize <= 8)
			mask = *decodeSrc++;
		else if (rowSize <= 16) {
			mask = VDReadUnalignedLEU16(decodeSrc2);
			decodeSrc2 += 2;
		} else if (rowSize <= 24) {
			mask = VDReadUnalignedLEU32(decodeSrc2) & 0xFFFFFF;
			decodeSrc2 += 3;
		} else {
			mask = VDReadUnalignedLEU32(decodeSrc2);
			decodeSrc2 += 4;
		}

		for(uint32 i=0; i<rowSize; ++i) {
			if (mask & (1 << i))
				decodeDst[i] = *decodeSrc2++;
		}

		const size_t srcUsed = decodeSrc2 - decodeSrc;
		if (srcEnd - src < (ptrdiff_t)srcUsed)
			throw ATInvalidSaveStateException();

		src += srcUsed;
	}

	if (src != srcEnd)
		throw ATInvalidSaveStateException();
}

////////////////////////////////////////////////////////////////////////////////

ATSavedTracePredictorXOR::ATSavedTracePredictorXOR(uint32 offset, uint32 size)
	: mXorOffset(offset)
	, mXorSize(size)
{
}

void ATSavedTracePredictorXOR::Validate(uint32 rowSize) const {
	if (mXorOffset >= rowSize || rowSize - mXorOffset < mXorSize)
		throw ATInvalidSaveStateException();
}

void ATSavedTracePredictorXOR::Reset() {
	memset(mPredBuf, 0, sizeof mPredBuf);
}

void ATSavedTracePredictorXOR::Encode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	dst += mXorOffset;

	const uint32 xorSize = mXorSize;
	uint8 *VDRESTRICT dst2 = dst;
	while(rowCount--) {
		for(uint32 i=0; i<xorSize; ++i) {
			uint8 c = dst2[i];
			dst2[i] = c ^ mPredBuf[i];
			mPredBuf[i] = c;
		}

		dst2 += rowSize;
	}
}

void ATSavedTracePredictorXOR::Decode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	dst += mXorOffset;

	const uint32 xorSize = mXorSize;
	uint8 *VDRESTRICT dst2 = dst;
	while(rowCount--) {
		for(uint32 i=0; i<xorSize; ++i) {
			uint8 v = mPredBuf[i] ^ dst2[i];
			mPredBuf[i] = v;
			dst2[i] = v;
		}

		dst2 += rowSize;
	}
}

////////////////////////////////////////////////////////////////////////////////

ATSavedTracePredictorEA::ATSavedTracePredictorEA(uint32 offset)
	: mOffset(offset)
{
}

void ATSavedTracePredictorEA::Validate(uint32 rowSize) const {
	if (mOffset >= rowSize || rowSize - mOffset < 4)
		throw ATInvalidSaveStateException();
}

void ATSavedTracePredictorEA::Reset() {
	mPrev = 0;
}

void ATSavedTracePredictorEA::Encode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	dst += mOffset;

	uint8 *VDRESTRICT dst2 = dst;
	uint32 prev = mPrev;
	while(rowCount--) {
		uint32 v = VDReadUnalignedLEU32(dst2);
		uint32 delta = v ^ prev;

		// -1 is common as it means no effective address, so we modify the encoding
		// to make the lower three bytes don't care in that case.
		if (v >= 0xFF000000)
			delta &= 0xFF000000;

		prev ^= delta;

		VDWriteUnalignedLEU32(dst2, delta);

		dst2 += rowSize;
	}

	mPrev = prev;
}

void ATSavedTracePredictorEA::Decode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	dst += mOffset;

	uint8 *VDRESTRICT dst2 = dst;
	uint32 prev = mPrev;

	while(rowCount--) {
		uint32 delta = VDReadUnalignedLEU32(dst2);
		prev ^= delta;

		uint32 v = prev;
		if (v >= 0xFF000000)
			v = 0xFFFFFFFF;

		VDWriteUnalignedLEU32(dst2, v);
		dst2 += rowSize;
	}

	mPrev = prev;
}

////////////////////////////////////////////////////////////////////////////////

ATSavedTracePredictorPC::ATSavedTracePredictorPC(uint32 offset)
	: mOffset(offset)
{
}

void ATSavedTracePredictorPC::Validate(uint32 rowSize) const {
	if (mOffset >= rowSize || rowSize - mOffset < 2)
		throw ATInvalidSaveStateException();
}

void ATSavedTracePredictorPC::Reset() {
	memset(mPredBuf, 0, sizeof mPredBuf);
	mPrevPC = 0;
}

void ATSavedTracePredictorPC::Encode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	dst += mOffset;

	uint8 *VDRESTRICT dst2 = dst;
	uint16 prevPC = mPrevPC;
	while(rowCount--) {
		uint16 pc = VDReadUnalignedLEU16(dst2);
		uint16 pcDelta = pc - (prevPC + 1);
		uint16& pred = mPredBuf[prevPC];
		
		uint16 encDelta = pcDelta - pred;
		pred = pcDelta;

		if (encDelta & 0x8000)
			encDelta ^= 0x7FFF;

		encDelta = (encDelta >> 15) | (encDelta << 1);

		VDWriteUnalignedLEU16(dst2, encDelta);

		prevPC = pc;
		dst2 += rowSize;
	}

	mPrevPC = prevPC;
}

void ATSavedTracePredictorPC::Decode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT dst2 = dst + mOffset;
	uint16 pc = mPrevPC;
	while(rowCount--) {
		uint16 encDelta = VDReadUnalignedLEU16(dst2);

		encDelta = (encDelta >> 1) | (encDelta << 15);

		if (encDelta & 0x8000)
			encDelta ^= 0x7FFF;

		uint16& pred = mPredBuf[pc];
		uint16 pcDelta = pred + encDelta;
		pred = pcDelta;

		pc += pcDelta + 1;

		VDWriteUnalignedLEU16(dst2, pc);
		dst2 += rowSize;
	}

	mPrevPC = pc;
}

////////////////////////////////////////////////////////////////////////////////

ATSavedTracePredictorInsn::ATSavedTracePredictorInsn(uint32 insnOffset, uint32 insnSize, uint32 pcOffset, uint32 flagsBitOffset)
	: mPCOffset(pcOffset)
	, mInsnOffset(insnOffset)
	, mInsnSize(insnSize)
	, mFlagsBitOffset(flagsBitOffset)
{
}

void ATSavedTracePredictorInsn::Validate(uint32 rowSize) const {
	if (mPCOffset >= rowSize || rowSize - mPCOffset < 2)
		throw ATInvalidSaveStateException();

	if (mInsnOffset >= rowSize || rowSize - mInsnOffset < 2)
		throw ATInvalidSaveStateException();

	const uint32 rowBitSize = rowSize*8;
	if (mFlagsBitOffset >= rowBitSize || rowBitSize - mFlagsBitOffset < mInsnSize)
		throw ATInvalidSaveStateException();

	// check for overlaps
	const uint32 flagsByteOffset = mFlagsBitOffset >> 3;

	if ((uint32)(flagsByteOffset - mPCOffset) < 2)
		throw ATInvalidSaveStateException();

	if ((uint32)(flagsByteOffset - mInsnOffset) < 2)
		throw ATInvalidSaveStateException();

	if (mPCOffset < mInsnOffset + 2 && mInsnOffset < mPCOffset + 2)
		throw ATInvalidSaveStateException();
}

void ATSavedTracePredictorInsn::Reset() {
	mFlagsOffset = mFlagsBitOffset >> 3;
	mFlagsShift = mFlagsBitOffset & 7;

	memset(mPredBuf, 0, sizeof mPredBuf);
	mPrevFlags = 0;
}

void ATSavedTracePredictorInsn::Encode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	const uint8 *VDRESTRICT dstPC = dst + mPCOffset;
	uint8 *VDRESTRICT dstInsn = dst + mInsnOffset;
	uint8 *VDRESTRICT dstFlags = dst + mFlagsOffset;

	while(rowCount--) {
		const uint16 pc = VDReadUnalignedLEU16(dstPC);

		// predict opcode based on PC
		uint8 flags = 0;

		for(uint32 i=0; i<mInsnSize; ++i) {
			uint8& pred = mPredBuf[(pc + i) & 0xFFFF];
			uint8 c = dstInsn[i];
			uint8 delta = c ^ pred;

			if (delta) {
				if (!c)
					flags |= 1 << i;
				else {
					pred = c;
					dstInsn[i] = delta;
				}
			} else {
				if (!c && (mPrevFlags & (1 << i)))
					flags |= 1 << i;

				dstInsn[i] = 0;
			}
		}
		
		mPrevFlags = flags;
		dstFlags[0] |= flags << mFlagsShift;

		dstPC += rowSize;
		dstInsn += rowSize;
		dstFlags += rowSize;
	}
}

void ATSavedTracePredictorInsn::Decode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	const uint8 *VDRESTRICT dstPC = dst + mPCOffset;
	uint8 *VDRESTRICT dstInsn = dst + mInsnOffset;
	const uint8 *VDRESTRICT dstFlags = dst + mFlagsOffset;

	while(rowCount--) {
		const uint16 pc = VDReadUnalignedLEU16(dstPC);
		const uint8 flags = dstFlags[0] >> mFlagsShift;

		for(uint32 i=0; i<mInsnSize; ++i) {
			uint8 c = dstInsn[i];
			uint8& VDRESTRICT pred = mPredBuf[(pc + i) & 0xFFFF];

			if (!(flags & (1 << i))) {
				c ^= pred;
				pred = c;
				dstInsn[i] = c;
			}
		}

		dstPC += rowSize;
		dstInsn += rowSize;
		dstFlags += rowSize;
	}
}

////////////////////////////////////////////////////////////////////////////////

ATSavedTracePredictorHorizDelta32::ATSavedTracePredictorHorizDelta32(uint32 dstOffset, uint32 predOffset)
	: mDstOffset(dstOffset)
	, mPredOffset(predOffset)
{
}

void ATSavedTracePredictorHorizDelta32::Validate(uint32 rowSize) const {
	if (mDstOffset >= rowSize || rowSize - mDstOffset < 4)
		throw ATInvalidSaveStateException();

	if (mPredOffset >= rowSize || rowSize - mPredOffset < 4)
		throw ATInvalidSaveStateException();

	if (mDstOffset < mPredOffset + 4 && mPredOffset < mDstOffset + 4)
		throw ATInvalidSaveStateException();
}

void ATSavedTracePredictorHorizDelta32::Reset() {
}

void ATSavedTracePredictorHorizDelta32::Encode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT dst2 = dst + mDstOffset;
	const uint8 *VDRESTRICT pred = dst + mPredOffset;

	while(rowCount--) {
		VDWriteUnalignedLEU32(dst2, VDReadUnalignedLEU32(dst2) - VDReadUnalignedLEU32(pred));

		dst2 += rowSize;
		pred += rowSize;
	}
}

void ATSavedTracePredictorHorizDelta32::Decode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT dst2 = dst + mDstOffset;
	const uint8 *VDRESTRICT pred = dst + mPredOffset;

	while(rowCount--) {
		VDWriteUnalignedLEU32(dst2, VDReadUnalignedLEU32(dst2) + VDReadUnalignedLEU32(pred));

		dst2 += rowSize;
		pred += rowSize;
	}
}

////////////////////////////////////////////////////////////////////////////////

ATSavedTracePredictorVertDelta32::ATSavedTracePredictorVertDelta32(uint32 offset, uint32 bias)
	: mOffset(offset)
	, mBias((sint32)bias)
{
}

void ATSavedTracePredictorVertDelta32::Validate(uint32 rowSize) const {
	if (mOffset >= rowSize || rowSize - mOffset < 4)
		throw ATInvalidSaveStateException();
}

void ATSavedTracePredictorVertDelta32::Reset() {
	mPrev = 0;
}

void ATSavedTracePredictorVertDelta32::Encode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT dst2 = dst + mOffset;

	while(rowCount--) {
		uint32 v = VDReadUnalignedLEU32(dst2);
		uint32 delta = v - (mPrev + (uint32)mBias);
		mPrev = v;

		VDWriteUnalignedLEU32(dst2, delta);

		dst2 += rowSize;
	}
}

void ATSavedTracePredictorVertDelta32::Decode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT dst2 = dst + mOffset;

	uint32 v = mPrev;
	while(rowCount--) {
		uint32 delta = VDReadUnalignedLEU32(dst2);

		v += delta + (uint32)mBias;
		VDWriteUnalignedLEU32(dst2, v);

		dst2 += rowSize;
	}

	mPrev = v;
}

////////////////////////////////////////////////////////////////////////////////

ATSavedTracePredictorVertDelta8::ATSavedTracePredictorVertDelta8(uint32 offset, uint32 count)
	: mOffset(offset)
	, mCount(count)
{
}

void ATSavedTracePredictorVertDelta8::Validate(uint32 rowSize) const {
	if (mOffset >= rowSize || rowSize - mOffset < mCount)
		throw ATInvalidSaveStateException();
}

void ATSavedTracePredictorVertDelta8::Reset() {
	memset(mPredBuf, 0, sizeof mPredBuf);
}

void ATSavedTracePredictorVertDelta8::Encode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT dst2 = dst + mOffset;

	while(rowCount--) {
		for(uint32 i=0; i<mCount; ++i) {
			uint8 v = dst2[i];
			dst2[i] = v - mPredBuf[i];
			mPredBuf[i] = v;
		}

		dst2 += rowSize;
	}
}

void ATSavedTracePredictorVertDelta8::Decode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT dst2 = dst + mOffset;

	while(rowCount--) {
		for(uint32 i=0; i<mCount; ++i) {
			uint8 c = mPredBuf[i] + dst2[i];
			mPredBuf[i] = c;
			dst2[i] = c;
		}

		dst2 += rowSize;
	}
}

////////////////////////////////////////////////////////////////////////////////

ATSavedTracePredictorDelta32TablePrev8::ATSavedTracePredictorDelta32TablePrev8(uint32 valueOffset, uint32 opcodeOffset)
	: mValueOffset(valueOffset)
	, mOpcodeOffset(opcodeOffset)
{
}

void ATSavedTracePredictorDelta32TablePrev8::Validate(uint32 rowSize) const {
	if (mValueOffset >= rowSize || rowSize - mValueOffset < 4)
		throw ATInvalidSaveStateException();

	if (mOpcodeOffset >= rowSize)
		throw ATInvalidSaveStateException();

	if ((uint32)(mOpcodeOffset - mValueOffset) < 4)
		throw ATInvalidSaveStateException();
}

void ATSavedTracePredictorDelta32TablePrev8::Reset() {
	memset(mPredBuf, 0, sizeof mPredBuf);
	mPrevOp = 0;
}

void ATSavedTracePredictorDelta32TablePrev8::Encode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT dst2 = dst + mValueOffset;
	const uint8 *VDRESTRICT opcode = dst + mOpcodeOffset;

	uint8 prevOp = mPrevOp;
	while(rowCount--) {
		uint32 v = VDReadUnalignedLEU32(dst2);

		VDWriteUnalignedLEU32(dst2, v - mPredBuf[prevOp]);
		mPredBuf[prevOp] = v;

		prevOp = *opcode;

		dst2 += rowSize;
		opcode += rowSize;
	}
	mPrevOp = prevOp;
}

void ATSavedTracePredictorDelta32TablePrev8::Decode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT dst2 = dst + mValueOffset;
	const uint8 *VDRESTRICT opcode = dst + mOpcodeOffset;

	uint8 prevOp = mPrevOp;
	while(rowCount--) {
		uint32 delta = VDReadUnalignedLEU32(dst2);
		uint32 v = mPredBuf[prevOp] + delta;

		mPredBuf[prevOp] = v;
		VDWriteUnalignedLEU32(dst2, v);

		prevOp = *opcode;

		dst2 += rowSize;
		opcode += rowSize;
	}
	mPrevOp = prevOp;
}

////////////////////////////////////////////////////////////////////////////////

ATSavedTracePredictorXor32Table8::ATSavedTracePredictorXor32Table8(uint32 valueOffset, uint32 predOffset)
	: mValueOffset(valueOffset)
	, mPredOffset(predOffset)
{
}

void ATSavedTracePredictorXor32Table8::Validate(uint32 rowSize) const {
	if (mValueOffset >= rowSize || rowSize - mValueOffset < 4)
		throw ATInvalidSaveStateException();

	if (mPredOffset >= rowSize)
		throw ATInvalidSaveStateException();

	if ((uint32)(mPredOffset - mValueOffset) < 4)
		throw ATInvalidSaveStateException();
}

void ATSavedTracePredictorXor32Table8::Reset() {
	memset(mPredBuf, 0, sizeof mPredBuf);
}

void ATSavedTracePredictorXor32Table8::Encode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT dst2 = dst + mValueOffset;
	const uint8 *VDRESTRICT pred = dst + mPredOffset;

	while(rowCount--) {
		uint32 v = VDReadUnalignedLEU32(dst2);
		uint8 p = *pred;

		VDWriteUnalignedLEU32(dst2, v ^ mPredBuf[p]);
		mPredBuf[p] = v;

		dst2 += rowSize;
		pred += rowSize;
	}
}

void ATSavedTracePredictorXor32Table8::Decode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT dst2 = dst + mValueOffset;
	const uint8 *VDRESTRICT pred = dst + mPredOffset;

	while(rowCount--) {
		uint32 delta = VDReadUnalignedLEU32(dst2);
		uint8 p = *pred;

		uint32 v = delta ^ mPredBuf[p];
		mPredBuf[p] = v;
		VDWriteUnalignedLEU32(dst2, v);

		dst2 += rowSize;
		pred += rowSize;
	}
}

////////////////////////////////////////////////////////////////////////////////

ATSavedTracePredictorXor32TablePrev16::ATSavedTracePredictorXor32TablePrev16(uint32 valueOffset, uint32 pcOffset)
	: mValueOffset(valueOffset)
	, mPCOffset(pcOffset)
{
}

void ATSavedTracePredictorXor32TablePrev16::Validate(uint32 rowSize) const {
	if (mValueOffset >= rowSize || rowSize - mValueOffset < 4)
		throw ATInvalidSaveStateException();

	if (mPCOffset >= rowSize || rowSize - mPCOffset < 2)
		throw ATInvalidSaveStateException();

	if (mValueOffset < mPCOffset + 2 && mPCOffset < mValueOffset + 4)
		throw ATInvalidSaveStateException();
}

void ATSavedTracePredictorXor32TablePrev16::Reset() {
	memset(mPredBuf, 0, sizeof mPredBuf);
	mPrevPC = 0;
}

void ATSavedTracePredictorXor32TablePrev16::Encode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT valPtr = dst + mValueOffset;
	const uint8 *VDRESTRICT pcPtr = dst + mPCOffset;
	
	uint16 prevPC = mPrevPC;
	while(rowCount--) {
		uint32 v = VDReadUnalignedU32(valPtr);

		VDWriteUnalignedU32(valPtr, v ^ mPredBuf[prevPC]);
		mPredBuf[prevPC] = v;

		prevPC = VDReadUnalignedLEU16(pcPtr);

		valPtr += rowSize;
		pcPtr += rowSize;
	}
	mPrevPC = prevPC;
}

void ATSavedTracePredictorXor32TablePrev16::Decode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT valPtr = dst + mValueOffset;
	const uint8 *VDRESTRICT pcPtr = dst + mPCOffset;

	uint16 prevPC = mPrevPC;
	while(rowCount--) {
		const uint32 delta = VDReadUnalignedU32(valPtr);
		uint32 v = mPredBuf[prevPC] ^ delta;

		mPredBuf[prevPC] = v;
		VDWriteUnalignedU32(valPtr, v);
		
		prevPC = VDReadUnalignedLEU16(pcPtr);

		valPtr += rowSize;
		pcPtr += rowSize;
	}
	mPrevPC = prevPC;
}

////////////////////////////////////////////////////////////////////////////////

ATSavedTracePredictorSignMag32::ATSavedTracePredictorSignMag32(uint32 offset)
	: mOffset(offset)
{
}

void ATSavedTracePredictorSignMag32::Validate(uint32 rowSize) const {
	if (mOffset >= rowSize || rowSize - mOffset < 4)
		throw ATInvalidSaveStateException();
}

void ATSavedTracePredictorSignMag32::Reset() {
}

void ATSavedTracePredictorSignMag32::Encode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT dst2 = dst + mOffset;

	while(rowCount--) {
		uint32 delta = VDReadUnalignedLEU32(dst2);

		if (delta & 0x80000000)
			delta ^= 0x7FFFFFFF;

		delta = (delta << 1) | (delta >> 31);

		VDWriteUnalignedLEU32(dst2, delta);

		dst2 += rowSize;
	}
}

void ATSavedTracePredictorSignMag32::Decode(uint8 *dst, uint32 rowSize, size_t rowCount) {
	uint8 *VDRESTRICT dst2 = dst + mOffset;

	while(rowCount--) {
		uint32 delta = VDReadUnalignedLEU32(dst2);

		delta = (delta << 31) | (delta >> 1);

		if (delta & 0x80000000)
			delta ^= 0x7FFFFFFF;

		VDWriteUnalignedLEU32(dst2, delta);

		dst2 += rowSize;
	}
}
